@@ -1186,7 +1186,11 @@ static const char * const ldst_name[] =
 #endif // _MSC_VER
 };
 
-void tcg_dump_ops(TCGContext *s)
+inline int minz(int a, int b) {
+    return a < b ? a : b;
+}
+
+void tcg_dump_ops(TCGContext *s, void *buffer, uint64_t buffer_size)
 {
     const uint16_t *opc_ptr;
     const TCGArg *args;
@@ -1196,6 +1200,8 @@ void tcg_dump_ops(TCGContext *s)
     const TCGOpDef *def;
     char buf[128];
 
+    int cx = 0;
+
     first_insn = 1;
     opc_ptr = s->gen_opc_buf;
     args = s->gen_opparam_buf;
@@ -1210,9 +1216,10 @@ void tcg_dump_ops(TCGContext *s)
             pc = args[0];
 #endif
             if (!first_insn) {
-                printf("\n");
+                cx += snprintf((char*)buffer + cx, buffer_size - minz(cx, buffer_size), "\n");
             }
-            printf(" ---- 0x%" PRIx64, pc);
+            cx += snprintf((char*)buffer + cx, buffer_size - minz(cx, buffer_size),
+                         " ---- 0x%" PRIx64, pc);
             first_insn = 0;
             nb_oargs = def->nb_oargs;
             nb_iargs = def->nb_iargs;
@@ -1227,12 +1234,13 @@ void tcg_dump_ops(TCGContext *s)
             nb_cargs = def->nb_cargs;
 
             /* function name, flags, out args */
-            printf(" %s %s,$0x%" TCG_PRIlx ",$%d", def->name,
+            cx += snprintf((char*)buffer + cx, buffer_size - minz(cx, buffer_size),
+                     " %s %s,$0x%" TCG_PRIlx ",$%d", def->name,
                      tcg_find_helper(s, args[nb_oargs + nb_iargs]),
                      args[nb_oargs + nb_iargs + 1], nb_oargs);
             for (i = 0; i < nb_oargs; i++) {
-                printf(",%s", tcg_get_arg_str_idx(s, buf, sizeof(buf),
-                                                   args[i]));
+                cx += snprintf((char*)buffer + cx, buffer_size - minz(cx, buffer_size), ",%s",
+                           tcg_get_arg_str_idx(s, buf, sizeof(buf), args[i]));
             }
             for (i = 0; i < nb_iargs; i++) {
                 TCGArg arg = args[nb_oargs + i];
@@ -1240,10 +1248,11 @@ void tcg_dump_ops(TCGContext *s)
                 if (arg != TCG_CALL_DUMMY_ARG) {
                     t = tcg_get_arg_str_idx(s, buf, sizeof(buf), arg);
                 }
-                printf(",%s", t);
+                cx += snprintf((char*)buffer + cx, buffer_size - minz(cx, buffer_size), ",%s", t);
             }
         } else {
-            printf(" %s ", def->name);
+            cx += snprintf((char*)buffer + cx, buffer_size - minz(cx, buffer_size), " %s ",
+                                 def->name);
             if (c == INDEX_op_nopn) {
                 /* variable number of arguments */
                 nb_cargs = *args;
@@ -1258,17 +1267,17 @@ void tcg_dump_ops(TCGContext *s)
             k = 0;
             for(i = 0; i < nb_oargs; i++) {
                 if (k != 0) {
-                    printf(",");
+                    cx += snprintf((char*)buffer + cx, buffer_size - minz(cx, buffer_size), ",");
                 }
-                printf("%s", tcg_get_arg_str_idx(s, buf, sizeof(buf),
-                                                   args[k++]));
+                cx += snprintf((char*)buffer + cx, buffer_size - minz(cx, buffer_size), "%s",
+                           tcg_get_arg_str_idx(s, buf, sizeof(buf), args[k++]));
             }
             for(i = 0; i < nb_iargs; i++) {
                 if (k != 0) {
-                    printf(",");
+                    cx += snprintf((char*)buffer + cx, buffer_size - minz(cx, buffer_size), ",");
                 }
-                printf("%s", tcg_get_arg_str_idx(s, buf, sizeof(buf),
-                                                   args[k++]));
+                cx += snprintf((char*)buffer + cx, buffer_size - minz(cx, buffer_size), "%s",
+                           tcg_get_arg_str_idx(s, buf, sizeof(buf), args[k++]));
             }
             switch (c) {
             case INDEX_op_brcond_i32:
@@ -1280,9 +1289,11 @@ void tcg_dump_ops(TCGContext *s)
             case INDEX_op_setcond_i64:
             case INDEX_op_movcond_i64:
                 if (args[k] < ARRAY_SIZE(cond_name) && cond_name[args[k]]) {
-                    printf(",%s", cond_name[args[k++]]);
+                    cx += snprintf((char*)buffer + cx, buffer_size - minz(cx, buffer_size),
+                                 ",%s", cond_name[args[k++]]);
                 } else {
-                    printf(",$0x%" TCG_PRIlx, args[k++]);
+                    cx += snprintf((char*)buffer + cx, buffer_size - minz(cx, buffer_size),
+                               ",$0x%" TCG_PRIlx, args[k++]);
                 }
                 i = 1;
                 break;
@@ -1291,10 +1302,12 @@ void tcg_dump_ops(TCGContext *s)
             case INDEX_op_qemu_ld_i64:
             case INDEX_op_qemu_st_i64:
                 if (args[k] < ARRAY_SIZE(ldst_name) && ldst_name[args[k]]) {
-                    printf(",%s", ldst_name[args[k++]]);
+                    cx += snprintf((char*)buffer + cx, buffer_size - minz(cx, buffer_size),
+                                 ",%s", ldst_name[args[k++]]);
                 } else {
-                    printf(",$0x%" TCG_PRIlx, args[k++]);
-                }
+                    cx += snprintf((char*)buffer + cx, buffer_size - minz(cx, buffer_size),
+                               ",$0x%" TCG_PRIlx, args[k++]);
+                 }
                 i = 1;
                 break;
             default:
@@ -1303,18 +1316,147 @@ void tcg_dump_ops(TCGContext *s)
             }
             for(; i < nb_cargs; i++) {
                 if (k != 0) {
-                    printf(",");
+                    cx += snprintf((char*)buffer + cx, buffer_size - minz(cx, buffer_size), ",");
                 }
                 arg = args[k++];
-                printf("$0x%" TCG_PRIlx, arg);
+                cx += snprintf((char*)buffer + cx, buffer_size - minz(cx, buffer_size),
+                               "$0x%" TCG_PRIlx, arg);
             }
         }
-        printf("\n");
+        cx += snprintf((char*)buffer + cx, buffer_size - minz(cx, buffer_size), "\n");
         args += nb_iargs + nb_oargs + nb_cargs;
     }
-    printf("###########\n");
 }
 
+// void tcg_dump_ops(TCGContext *s)
+// {
+//     const uint16_t *opc_ptr;
+//     const TCGArg *args;
+//     TCGArg arg;
+//     TCGOpcode c;
+//     int i, k, nb_oargs, nb_iargs, nb_cargs, first_insn;
+//     const TCGOpDef *def;
+//     char buf[128];
+
+//     first_insn = 1;
+//     opc_ptr = s->gen_opc_buf;
+//     args = s->gen_opparam_buf;
+//     while (opc_ptr < s->gen_opc_ptr) {
+//         c = *opc_ptr++;
+//         def = &s->tcg_op_defs[c];
+//         if (c == INDEX_op_debug_insn_start) {
+//             uint64_t pc;
+// #if TARGET_LONG_BITS > TCG_TARGET_REG_BITS
+//             pc = ((uint64_t)args[1] << 32) | args[0];
+// #else
+//             pc = args[0];
+// #endif
+//             if (!first_insn) {
+//                 printf("\n");
+//             }
+//             printf(" ---- 0x%" PRIx64, pc);
+//             first_insn = 0;
+//             nb_oargs = def->nb_oargs;
+//             nb_iargs = def->nb_iargs;
+//             nb_cargs = def->nb_cargs;
+//         } else if (c == INDEX_op_call) {
+//             TCGArg arg;
+
+//             /* variable number of arguments */
+//             arg = *args++;
+//             nb_oargs = arg >> 16;
+//             nb_iargs = arg & 0xffff;
+//             nb_cargs = def->nb_cargs;
+
+//             /* function name, flags, out args */
+//             printf(" %s %s,$0x%" TCG_PRIlx ",$%d", def->name,
+//                      tcg_find_helper(s, args[nb_oargs + nb_iargs]),
+//                      args[nb_oargs + nb_iargs + 1], nb_oargs);
+//             for (i = 0; i < nb_oargs; i++) {
+//                 printf(",%s", tcg_get_arg_str_idx(s, buf, sizeof(buf),
+//                                                    args[i]));
+//             }
+//             for (i = 0; i < nb_iargs; i++) {
+//                 TCGArg arg = args[nb_oargs + i];
+//                 const char *t = "<dummy>";
+//                 if (arg != TCG_CALL_DUMMY_ARG) {
+//                     t = tcg_get_arg_str_idx(s, buf, sizeof(buf), arg);
+//                 }
+//                 printf(",%s", t);
+//             }
+//         } else {
+//             printf(" %s ", def->name);
+//             if (c == INDEX_op_nopn) {
+//                 /* variable number of arguments */
+//                 nb_cargs = *args;
+//                 nb_oargs = 0;
+//                 nb_iargs = 0;
+//             } else {
+//                 nb_oargs = def->nb_oargs;
+//                 nb_iargs = def->nb_iargs;
+//                 nb_cargs = def->nb_cargs;
+//             }
+
+//             k = 0;
+//             for(i = 0; i < nb_oargs; i++) {
+//                 if (k != 0) {
+//                     printf(",");
+//                 }
+//                 printf("%s", tcg_get_arg_str_idx(s, buf, sizeof(buf),
+//                                                    args[k++]));
+//             }
+//             for(i = 0; i < nb_iargs; i++) {
+//                 if (k != 0) {
+//                     printf(",");
+//                 }
+//                 printf("%s", tcg_get_arg_str_idx(s, buf, sizeof(buf),
+//                                                    args[k++]));
+//             }
+//             switch (c) {
+//             case INDEX_op_brcond_i32:
+//             case INDEX_op_setcond_i32:
+//             case INDEX_op_movcond_i32:
+//             case INDEX_op_brcond2_i32:
+//             case INDEX_op_setcond2_i32:
+//             case INDEX_op_brcond_i64:
+//             case INDEX_op_setcond_i64:
+//             case INDEX_op_movcond_i64:
+//                 if (args[k] < ARRAY_SIZE(cond_name) && cond_name[args[k]]) {
+//                     printf(",%s", cond_name[args[k++]]);
+//                 } else {
+//                     printf(",$0x%" TCG_PRIlx, args[k++]);
+//                 }
+//                 i = 1;
+//                 break;
+//             case INDEX_op_qemu_ld_i32:
+//             case INDEX_op_qemu_st_i32:
+//             case INDEX_op_qemu_ld_i64:
+//             case INDEX_op_qemu_st_i64:
+//                 if (args[k] < ARRAY_SIZE(ldst_name) && ldst_name[args[k]]) {
+//                     printf(",%s", ldst_name[args[k++]]);
+//                 } else {
+//                     printf(",$0x%" TCG_PRIlx, args[k++]);
+//                 }
+//                 i = 1;
+//                 break;
+//             default:
+//                 i = 0;
+//                 break;
+//             }
+//             for(; i < nb_cargs; i++) {
+//                 if (k != 0) {
+//                     printf(",");
+//                 }
+//                 arg = args[k++];
+//                 printf("$0x%" TCG_PRIlx, arg);
+//             }
+//         }
+//         printf("\n");
+//         args += nb_iargs + nb_oargs + nb_cargs;
+//     }
+//     printf("###########\n");
+// }
+
 /* we give more priority to constraints with less registers */
 static int get_constraint_priority(const TCGOpDef *def, int k)
 {
@@ -2537,7 +2679,7 @@ static inline int tcg_gen_code_common(TCGContext *s,
 #ifdef DEBUG_DISAS
     if (unlikely(qemu_loglevel_mask(CPU_LOG_TB_OP))) {
         qemu_log("OP:\n");
-        tcg_dump_ops(s);
+        // tcg_dump_ops(s);
         qemu_log("\n");
     }
 #endif
@@ -2569,7 +2711,7 @@ static inline int tcg_gen_code_common(TCGContext *s,
 #ifdef DEBUG_DISAS
     if (unlikely(qemu_loglevel_mask(CPU_LOG_TB_OP_OPT))) {
         qemu_log("OP after optimization and liveness analysis:\n");
-        tcg_dump_ops(s);
+        // tcg_dump_ops(s);
         qemu_log("\n");
     }
 #endif
